﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Threading;
using System.Reflection;
using System.IO;
using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;
using System.Threading.Tasks;

namespace RCAddInCentroidCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
        const string CentroidCommandName = "RootPro.AddIn.CentroidCommand";     //コマンド名

        // 図心クラスのリスト
        List<ShapeCentroid> shapeCentroids;

        // 曲線の分割数
        int divideCount = 256;

        /// <summary>
        /// アドイン起動時に呼ばれる処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            // 多言語対応リソースのDLLファイルをロード
            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

            // コマンドの追加
            CommandManager.AddMacroCommand("RCAddInCentroidCS", Properties.Resources.CommandName, MacroCommand);

            Command command = CommandManager.AddCommand(CentroidCommandName);
            command.Name = Properties.Resources.CommandName;

            // 第1パラメーター「図形」
            command.ParameterGroups.Add(ParameterItemType.ShapeGroup, 0, Properties.Resources.ShapesParameterName);
            ShapeGroupParameterItem shapeGroupParameterItem = (ShapeGroupParameterItem)command.GetParameterItemByID(0);
            shapeGroupParameterItem.Description = Properties.Resources.ShapesParameterDescription;
            shapeGroupParameterItem.EnableBeforeSelection = false;
            shapeGroupParameterItem.ShapeTypeFilter =
                ShapeTypeFilters.Line | ShapeTypeFilters.Polyline | ShapeTypeFilters.Circle | ShapeTypeFilters.Arc |
                ShapeTypeFilters.Ellipse | ShapeTypeFilters.EllipseArc | ShapeTypeFilters.BSplineCurve;

            // 第2パラメーター「曲線の分割数」
            command.ParameterGroups.Add(ParameterItemType.Integer, 1, Properties.Resources.DivisionNumberParameterName);
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(1);
            integerParameterItem.Description = Properties.Resources.DivisionNumberParameterDescription;
            integerParameterItem.Value = 256;
            integerParameterItem.ValueSaveMode = ParameterValueSaveMode.Always; // パラメータ値の保存方法

            // 第3パラメーター「動作の種類」
            command.ParameterGroups.Add(ParameterItemType.List, 2, Properties.Resources.OperationTypeParameterName);
            ListParameterItem listParameterItem = (ListParameterItem)command.GetParameterItemByID(2); // パラメータアイテム
            string[] groupString = { Properties.Resources.OperationTypeParameterListItem1, Properties.Resources.OperationTypeParameterListItem2 };
            listParameterItem.Items = groupString;
            listParameterItem.SetEmpty();
            listParameterItem.Description = Properties.Resources.OperationTypeParameterDescription;
            listParameterItem.ValueSaveMode = ParameterValueSaveMode.OneTime;

            // 開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(Command_Begin);
            command.End += new EventHandler(Command_End);
        }

        /// <summary>
        /// アドイン終了時に呼ばれる処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            CommandManager.RemoveCommand(CentroidCommandName);
            CommandManager.RemoveMacroCommand(MacroCommand);
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        /// <summary>
        /// 外部フォルダのアセンブリを動的にロード
        /// 多言語対応リソースのDLLファイルをロード
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
        {
            AssemblyName assemblyName = new AssemblyName(args.Name);

            foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
            {
                if(AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
                    return assembly;
            }

            if(assemblyName.Name.EndsWith(".resources"))
            {
                if(assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
                {
                    string dllPath = Path.GetDirectoryName(this.AddInFullPath);
                    dllPath = Path.Combine(dllPath, assemblyName.CultureName);
                    dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

                    if(File.Exists(dllPath))
                    {
                        try
                        {
                            return Assembly.LoadFrom(dllPath);
                        }
                        catch
                        {
                        }
                    }
                }
            }

            return null;
        }


        /// <summary>
        /// マクロコマンドの実行
        /// </summary>
        private void MacroCommand()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(CentroidCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            shapeCentroids = new List<ShapeCentroid>();

            // 「動作の種類」パラメータを空にするため一旦削除して追加する
            command.ParameterGroups.RemoveAt(2);
            command.ParameterGroups.Add(ParameterItemType.List, 2, Properties.Resources.OperationTypeParameterName);
            ListParameterItem listParameterItem = (ListParameterItem)command.GetParameterItemByID(2);
            string[] groupString = { Properties.Resources.OperationTypeParameterListItem1, Properties.Resources.OperationTypeParameterListItem2 };
            listParameterItem.Items = groupString;
            listParameterItem.SetEmpty();
            listParameterItem.Description = Properties.Resources.OperationTypeParameterDescription;
            listParameterItem.ValueSaveMode = ParameterValueSaveMode.InCommand;

            //パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(Command_ParameterChanged);

            //マウス移動イベントの追加
            command.MouseMove += new CommandMouseMoveEventHandler(Command_MouseMove);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(Command_ParameterChanged);

            //マウス移動イベントの削除
            command.MouseMove -= new CommandMouseMoveEventHandler(Command_MouseMove);
        }

        /// <summary>
        /// マウス移動イベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Command_MouseMove(Object sender, CommandMouseMoveEventArgs e)
        {
            Command command = (Command)sender;

            // 図心オブジェクトがない場合
            if(shapeCentroids.Count == 0)
                return;

            // 図心オブジェクトリストの数だけループ
            for(int i = 0; i < shapeCentroids.Count; i++)
            {
                try
                {
                    // ラバーバンド図形がない場合
                    if(shapeCentroids[i].rubberBandShapes.Count == 0)
                    {
                        // ラバーバンド図形を作成
                        shapeCentroids[i].CreateRubberBandShapes();
                    }
                    else
                    {
                        // ラバーバンド描画
                        command.DrawShapeRubberBand(shapeCentroids[i].rubberBandShapes.ToArray());
                    }
                }
                catch
                {
                    return;
                }
            }
        }

        /// <summary>
        /// パラメータ変更イベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            Document doc = ActiveDocument;         //ドキュメントの取得

            Command command = (Command)sender;
            int id = e.ParameterID;// 変更のあったパラメータID

            // 図形が変更された場合
            if(id == 0)
            {
                // 図形数のチェック
                ShapeGroupParameterItem param = (ShapeGroupParameterItem)command.GetParameterItemByID(0);
                if(param.SelectedShapes.Count > 1000)
                {
                    e.CommandState = CommandState.ParameterError;
                    return;
                }

                // 図心の結果をクリア
                shapeCentroids.Clear();

                if(param.IsEmpty)
                    return;
            }

            // 分割数が変更された場合
            if(id == 1)
            {
                IntegerParameterItem param = (IntegerParameterItem)command.GetParameterItemByID(1);
                if(!param.IsEmpty && (param.Value < 10 || param.Value > 10000))
                {
                    e.CommandState = CommandState.ParameterError;
                    return;
                }
                // 点列の情報をクリア
                foreach(var sc in shapeCentroids)
                    sc.ClearPoints();

                if(param.IsEmpty)
                    return;
            }

            // 図形を取得
            ShapeGroupParameterItem shapeGroupParameterItem = (ShapeGroupParameterItem)command.GetParameterItemByID(0);
            if(shapeGroupParameterItem.SelectedShapes.Count == 0 || shapeGroupParameterItem.IsEmpty)
                return;

            // 分割数を取得
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(1);
            if(integerParameterItem.IsEmpty)
                return;
            divideCount = integerParameterItem.Value;

            // 図心未計算の場合
            if(shapeCentroids.Count == 0)
            {
                // 他の図形と端点でつながる可能性のある図形のリスト
                List<SelectedShape> connectableShapes = new List<SelectedShape>();

                // 図形を列挙する
                var shapes = shapeGroupParameterItem.SelectedShapes;
                foreach(var selectedShape in shapes)
                {
                    var shape = selectedShape.Shape;
                    if(shape is LineShape || shape is ArcShape || shape is EllipseArcShape ||
                       (shape is PolylineShape p && !p.IsClosed) ||
                       (shape is BSplineCurveShape bs && !bs.IsClosed))
                    {
                        connectableShapes.Add(selectedShape);
                    }
                    // 円、楕円、閉じた連続線、閉じたBスプライン
                    else
                    {
                        // 閉じた図形は図形単体で図心計算するため、図心クラスのオブジェクトを作成してリストに追加する
                        shapeCentroids.Add(new ShapeCentroid(this, new List<Shape> { selectedShape.Shape }, selectedShape.Layout));
                    }
                }

                // リストの複製
                List<SelectedShape> copyShapes = new List<SelectedShape>(connectableShapes);
                // 接続可能な図形を列挙
                foreach(var selectedShape in connectableShapes)
                {
                    // すでに接続済みか
                    if(!copyShapes.Contains(selectedShape))
                        continue;
                    else
                    {
                        copyShapes.Remove(selectedShape);

                        List<Shape> connectedShapes = new List<Shape> { selectedShape.Shape };
                        Layout layout = selectedShape.Layout;

                        // 図形の端点座標を取得
                        if(GetShapeEndPoint(selectedShape.Shape, out Point2d p1, out Point2d p2))
                        {
                            // shape に接続される図形を再帰的に検索する
                            ConnectShapes(p1, p2, layout, ref copyShapes, ref connectedShapes);
                        }

                        // 図心クラスのオブジェクトを作成してリストに追加する
                        shapeCentroids.Add(new ShapeCentroid(this, connectedShapes, layout));
                    }
                }
            }

            // 図心を計算する
            for(int i = 0; i < shapeCentroids.Count; i++)
                shapeCentroids[i].Calculate(divideCount);

            // 動作の種類
            ListParameterItem listParameterItem = (ListParameterItem)command.GetParameterItemByID(2);
            if(listParameterItem.IsEmpty)
                return;

            int count = 0; 
            // 選択中の動作の種類を取得
            int selectedIndex = listParameterItem.SelectedIndex;

            // 座標値をクリップボードにコピー
            if(selectedIndex == 0)
            {
                // 図心計算の結果を待つ
                WaitCalculation();

                // テキストデータ作成
                string copyText = "";
                for(int i = 0; i < shapeCentroids.Count; i++)
                {
                    // 自己交差のチェック
                    if(shapeCentroids[i].HasSelfIntersection())
                        continue;

                    // 有効な領域かチェック
                    if(!shapeCentroids[i].IsValidRegion())
                        continue;

                    if(shapeCentroids[i].GetCentroid(out Point2d centroid))
                    {
                        copyText += centroid.ToString() + "\r\n";
                        shapeCentroids[i].layout.Drawing.Shapes.AddPointMarker(centroid);
                        count++;
                    }
                }

                if(count != 0)
                {
                    // クリップボードにコピー
                    Clipboard.SetText(copyText);
                    StatusBar.SetInformation(Properties.Resources.ClipboardCopiedMsg, 3000);
                }
                else
                {
                    StatusBar.SetInformation(Properties.Resources.InvalidShapesMsg, 3000);
                }
            }
            // 点図形を作成
            else if(selectedIndex == 1)
            {
                // 図心計算の結果を待つ
                WaitCalculation();

                // UNDO 可能な操作を開始
                doc.UndoManager.BeginUndoUnit();

                for(int i = 0; i < shapeCentroids.Count; i++)
                {
                    // 自己交差のチェック
                    if(shapeCentroids[i].HasSelfIntersection())
                        continue;

                    // 有効な領域かチェック
                    if(!shapeCentroids[i].IsValidRegion())
                        continue;

                    // 既定値で点図形を作成する
                    if(shapeCentroids[i].GetCentroid(out Point2d centroid))
                    {
                        shapeCentroids[i].layout.Drawing.Shapes.AddPointMarker(centroid);
                        count++;
                    }
                }

                // UNDO 可能な操作を終了
                doc.UndoManager.EndUndoUnit();

                if(count != 0)
                    StatusBar.SetInformation(Properties.Resources.PointMarkerCreatedMsg, 3000);
                else
                    StatusBar.SetInformation(Properties.Resources.InvalidShapesMsg, 3000);
            }
            else
            {
                return;
            }

            e.CommandState = CommandState.End;
        }

        /// <summary>
        /// 端点が一致する図形を再帰的に検索。接続された図形のリストを返す
        /// </summary>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        /// <param name="searchShapes"></param>
        /// <param name="connectedShapes"></param>
        /// <param name="closed"></param>
        /// <returns>図形数</returns>
        private int ConnectShapes(in Point2d point1, in Point2d point2, in Layout layout, ref List<SelectedShape> searchShapes, ref List<Shape> connectedShapes)
        {
            if(connectedShapes.Count < 1)
                return 0;

            // 端点が一致した図形
            SelectedShape findShape;
            bool bFind = false;

            // 接続された後の新しい端点
            Point2d newPoint1 = new Point2d(0, 0);
            Point2d newPoint2 = new Point2d(0, 0);

            // 検索図形を列挙
            foreach(var selectedShape in searchShapes)
            {
                // レイアウトが同じ図形のみ対象とする
                if(selectedShape.Layout != layout)
                    continue;

                // 図形の端点座標を取得
                if(GetShapeEndPoint(selectedShape.Shape, out Point2d p1, out Point2d p2))
                {
                    // いずれかの端点と一致する場合
                    if(Geometry.IsPointEquals(point1, p1) || Geometry.IsPointEquals(point1, p2) ||
                        Geometry.IsPointEquals(point2, p1) || Geometry.IsPointEquals(point2, p2))
                    {
                        // 接続点を取得
                        Point2d connectPoint = Geometry.IsPointEquals(point1, p1) || Geometry.IsPointEquals(point1, p2) ? point1 : point2;

                        // 元の点列の始点側に連結
                        if(Geometry.IsPointEquals(connectPoint, point1))
                        {
                            // 接続していない方の点を取得
                            newPoint1 = Geometry.IsPointEquals(connectPoint, p1) ? p2 : p1;
                            newPoint2 = point2;
                        }
                        // 元の点列の終点側に連結
                        else
                        {
                            newPoint1 = point1;
                            // 接続していない方の点を取得
                            newPoint2 = Geometry.IsPointEquals(connectPoint, p1) ? p2 : p1;
                        }

                        findShape = selectedShape;
                        bFind = true;
                        break;
                    }
                }
            }

            // 端点が一致する図形が見つかった場合
            if(bFind)
            {
                // 検索リストから外す
                searchShapes.Remove(findShape);

                // 接続図形に追加
                connectedShapes.Add(findShape.Shape);

                // 閉じていない場合
                if(!Geometry.IsPointEquals(newPoint1, newPoint2))
                    // さらに接続する図形を探す
                    ConnectShapes(newPoint1, newPoint2, layout, ref searchShapes, ref connectedShapes);
            }

            return connectedShapes.Count;
        }

        /// <summary>
        /// 図形の端点の座標を返す
        /// </summary>
        /// <param name="shape"></param>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        private bool GetShapeEndPoint(in Shape shape, out Point2d point1, out Point2d point2)
        {
            bool result = false;
            point1 = new Point2d(0, 0);
            point2 = new Point2d(0, 0);

            // 線分
            if(shape is LineShape line)
            {
                // 始点、終点
                point1 = line.StartPoint;
                point2 = line.EndPoint;
                result = true;
            }
            // 円弧
            else if(shape is ArcShape arc)
            {
                // 始点、終点
                point1 = arc.StartPoint;
                point2 = arc.EndPoint;
                result = true;
            }
            // 楕円弧
            else if(shape is EllipseArcShape ellipseArc)
            {
                // 始点、終点
                point1 = ellipseArc.StartPoint;
                point2 = ellipseArc.EndPoint;
                result = true;
            }
            // 連続線
            else if(shape is PolylineShape polyline)
            {
                if(!polyline.IsClosed)
                {
                    // 点列の先頭と最後
                    point1 = polyline.Points[0];
                    point2 = polyline.Points[polyline.PointCount - 1];
                    result = true;
                }
            }
            // Bスプライン
            else if(shape is BSplineCurveShape bspline)
            {
                if(!bspline.IsClosed)
                {
                    // 開始パラメータ、終了パラメータの座標
                    point1 = bspline.GetPointAtParameter(bspline.StartParameter);
                    point2 = bspline.GetPointAtParameter(bspline.EndParameter);
                    result = true;
                }
            }

            return result;
        }

        /// <summary>
        /// 図心の計算結果を待つ
        /// </summary>
        /// <param name="shape"></param>
        /// <param name="point1"></param>
        /// <param name="point2"></param>
        private void WaitCalculation()
        {
            while(true)
            {
                bool result = true;

                for(int i = 0; i < shapeCentroids.Count; i++)
                {
                    if(shapeCentroids[i].HasSelfIntersection() || !shapeCentroids[i].IsValidRegion())
                        continue;

                    if((result &= shapeCentroids[i].centroidCalcFlg) == false)
                        break;
                }

                if(result)
                    return;

                Thread.Sleep(100);
            }
        }
     
        /// <summary>
        /// 図心クラス
        /// </summary>
        class ShapeCentroid
        {
            readonly private ApplicationEntryPoint app;

            // 図心
            public Point2d centroid;

            // 図心を算出する点列
            public List<Point2d> calcPoints = null;

            // 自己交差しているか
            private bool hasIntersection = false;

            // 自己交差チェック済みフラグ
            private bool intersectionCheckFlg = false;

            // 領域チェック結果
            private bool isValidRegion = false;

            // 領域チェック済みフラグ
            private bool validRegionCheckFlg = false;

            // 図心計算済みフラグ
            public bool centroidCalcFlg = false;

            // 端点で接続された図形のリスト
            private List<Shape> shapes = null;

            // 選択された図形のレイアウト
            public Layout layout;

            // ラバーバンド図形
            public List<Shape> rubberBandShapes = null;

            // ラバーバンド図形作成済みフラグ
            private bool rubberBandFlg = false;

            /// <summary>
            /// コンストラクタ
            /// </summary>
            public ShapeCentroid(ApplicationEntryPoint app, List<Shape> shapes, Layout layout)
            {
                this.centroid = new Point2d();
                this.calcPoints = new List<Point2d>();
                this.rubberBandShapes = new List<Shape>();

                this.app = app;
                this.shapes = new List<Shape>(shapes);
                this.layout = layout;
            }

            /// <summary>
            /// 自己交差しているか
            /// </summary>
            /// <returns></returns>
            public bool HasSelfIntersection()
            {
                if(calcPoints.Count == 0)
                    return false;

                // 1回のみチェックする
                if(!intersectionCheckFlg)
                {
                    intersectionCheckFlg = true;

                    PolylineShape polylineShape = app.ShapeFactory.CreatePolyline(calcPoints.ToArray());
                    hasIntersection = polylineShape.HasSelfIntersection;
                }

                return hasIntersection;
            }

            /// <summary>
            /// 有効な領域か
            /// </summary>
            /// <returns></returns>
            public bool IsValidRegion()
            {
                if(calcPoints.Count < 4)
                    return false;

                // 1回のみチェックする
                if(!validRegionCheckFlg)
                {
                    validRegionCheckFlg = true;

                    // 最初の二点で無限線を作成する
                    double angle = app.Geometry.GetAngle(calcPoints[0], calcPoints[1]);
                    var infiniteLine = app.ShapeFactory.CreateInfiniteLine(calcPoints[0], angle);
                    for(int i = 2; i < calcPoints.Count; i++)
                    {
                        // 全ての点が同一線上でなければtrue
                        if(!infiniteLine.IsPointOnShape(calcPoints[i]))
                        {
                            isValidRegion = true;
                            break;
                        }
                    }
                }

                return isValidRegion;
            }

            /// <summary>
            /// 図心計算
            /// </summary>
            /// <param name="divideCount"></param>
            public async void Calculate(int divideCount)
            {
                // 未計算の場合
                if(!centroidCalcFlg)
                {
                    // 図形を点列にする
                    if(await Task.Run(() => CalculatePoints(divideCount)).ConfigureAwait(false))
                    {
                        // 点列から図心を計算
                        if(await Task.Run(() => CalculateCentroid()).ConfigureAwait(false))
                            centroidCalcFlg = true;
                    }
                }
            }

            /// <summary>
            /// 図形を指定した分割数で分割する
            /// </summary>
            /// <returns></returns>
            public bool CalculatePoints(int divideCount)
            {
                // 図形を分割した点列を取得する
                GetShapePoints(shapes[0], divideCount, out List<Point2d> shapePoints);

                for(int i = 1; i < shapes.Count; i++)
                {
                    Point2d point1 = shapePoints[0];
                    Point2d point2 = shapePoints[shapePoints.Count - 1];

                    // 図形を分割した点列を取得する
                    GetShapePoints(shapes[i], divideCount, out List<Point2d> points);

                    Point2d p1 = points[0];
                    Point2d p2 = points[points.Count - 1];

                    // 接続点を取得
                    Point2d connectPoint = app.Geometry.IsPointEquals(point1, p1) || app.Geometry.IsPointEquals(point1, p2) ? point1 : point2;
                    // 元の点列の始点側に連結
                    if(app.Geometry.IsPointEquals(connectPoint, point1))
                    {
                        // 向きが逆の場合は点列を反転
                        if(app.Geometry.IsPointEquals(connectPoint, p1))
                            points.Reverse();

                        // 接続点を点列から削除
                        points.RemoveAt(points.Count - 1);

                        shapePoints.InsertRange(0, points);
                    }
                    // 元の点列の終点側に連結
                    else
                    {
                        // 向きが逆の場合は点列を反転
                        if(app.Geometry.IsPointEquals(connectPoint, p2))
                            points.Reverse();

                        // 接続点を点列から削除
                        points.RemoveAt(0);

                        shapePoints.AddRange(points);
                    }
                }

                // 閉じていない場合は閉じる
                if(!app.Geometry.IsPointEquals(shapePoints[0], shapePoints[shapePoints.Count - 1]))
                    shapePoints.Add(shapePoints[0]);

                // 点列を保存
                this.calcPoints = new List<Point2d>(shapePoints);

                return true;
            }

            /// <summary>
            /// 点列から図心を計算する
            /// </summary>
            /// <returns></returns>
            public bool CalculateCentroid()
            {
                // 自己交差チェック
                if(HasSelfIntersection())
                    return false;

                int pointCount = calcPoints.Count;
                if(pointCount < 4)
                {
                    return false;
                }

                double area2 = 0.0;
                double sumX = 0.0;
                double sumY = 0.0;

                for(int i = 0; i < pointCount - 1; i++)
                {
                    Point2d p1 = calcPoints[i];
                    Point2d p2 = calcPoints[i + 1];

                    // 面積計算用
                    double addValue = (p1.X * p2.Y - p2.X * p1.Y);
                    area2 += addValue;

                    // 図心計算用
                    double addValueX = (p1.X + p2.X) * (p1.X * p2.Y - p2.X * p1.Y);
                    double addValueY = (p1.Y + p2.Y) * (p1.X * p2.Y - p2.X * p1.Y);
                    sumX += addValueX;
                    sumY += addValueY;
                }

                double area = area2 * 0.5;

                centroid.X = sumX / (6.0 * area);
                centroid.Y = sumY / (6.0 * area);

                return true;
            }

            /// <summary>
            /// 図心を取得する
            /// </summary>
            /// <param name="point"></param>
            public bool GetCentroid(out Point2d point)
            {
                if(!centroidCalcFlg)
                {
                    point = new Point2d();
                    return false;
                }

                point = centroid;
                return true;
            }

            /// <summary>
            /// 図形を分割した点列を返す
            /// </summary>
            /// <param name="shape"></param>
            /// <param name="points"></param>
            private int GetShapePoints(in Shape shape, int divideCount, out List<Point2d> points)
            {
                points = new List<Point2d>();

                // 線分
                if(shape is LineShape line)
                {
                    // 始点、終点
                    points.Add(line.StartPoint);
                    points.Add(line.EndPoint);
                }
                // 連続線
                else if(shape is PolylineShape polyline)
                {
                    points.AddRange(new List<Point2d>(polyline.Points));
                }
                // 円弧、楕円弧、Bスプライン
                else
                {
                    if(shape is BasicShape basicShape)
                    {
                        double resolution = basicShape.GetLength(basicShape.StartParameter, basicShape.EndParameter) / divideCount;

                        // 連続線に変換
                        PolylineShape polyline2 = basicShape.ConvertToPolyline(resolution, divideCount);
                        // 点列を追加
                        points.AddRange(new List<Point2d>(polyline2.Points));
                    }
                }

                return points.Count;
            }

            /// <summary>
            /// 図心の点列をクリア
            /// </summary>
            public bool ClearPoints()
            {
                // 点列クリア
                calcPoints.Clear();

                // 自己交差チェック
                hasIntersection = false;
                // 自己交差チェック済みフラグ
                intersectionCheckFlg = false;

                // 領域チェック
                isValidRegion = false;
                // 領域チェック済みフラグ
                validRegionCheckFlg = false;

                // 図心計算済みフラグ
                centroidCalcFlg = false;

                // ラバーバンド図形
                rubberBandShapes.Clear();

                // ラバーバンド図形作成済みフラグ
                rubberBandFlg = false;

                return true;
            }

            /// <summary>
            /// ラバーバンド図形を非同期で作成
            /// </summary>
            /// <returns></returns>

            private async void CreateRubberBandShapesAsync()
            {
                rubberBandFlg = true;

                rubberBandShapes = await Task.Run(() =>
                {
                    List<Shape> shapes = new List<Shape>();

                    var doc = app.ActiveDocument;
                    var shapeFactory = app.ShapeFactory;

                    Point2d[] ps = calcPoints.ToArray();
                    // 図心を計算する範囲をハッチングのラバーバンドで表示する
                    var polyline = shapeFactory.CreatePolyline(ps);
                    HatchShape hatch = shapeFactory.CreateHatch(new Shape[] { polyline });

                    hatch.HatchType = HatchType.Solid;
                    hatch.BackgroundColorNumber = 3;
                    hatch.Transparency = 0.8f;

                    // 現在の座標系に変換
                    Shape hatch2 = layout.TransformShapeLocalToWorld(hatch);
                    Shape hatch3 = doc.CurrentLayout.TransformShapeWorldToLocal(hatch2);

                    // ラバーバンド図形に追加
                    shapes.Add(hatch3);

                    // 図心位置を点のラバーバンドで表示する
                    var pointMarker = shapeFactory.CreatePointMarker(centroid);

                    // 既定値をコピー
                    doc.PointMarkerSettings.CopyToShape(pointMarker);
                    if(pointMarker.MarkerType == PointMarkerType.Dot)
                        pointMarker.MarkerType = PointMarkerType.X;
                    pointMarker.ColorNumber = 2;

                    // 現在の座標系に変換
                    Shape pointMarker2 = layout.TransformShapeLocalToWorld(pointMarker);
                    Shape pointMarker3 = doc.CurrentLayout.TransformShapeWorldToLocal(pointMarker2);

                    // ラバーバンド図形に追加
                    shapes.Add(pointMarker3);

                    return shapes;
                });
           }

            /// <summary>
            /// ラバーバンド図形を作成
            /// </summary>
            /// <returns></returns>
            public void CreateRubberBandShapes()
            {
                if(!centroidCalcFlg)
                    return;

                if(HasSelfIntersection() || !IsValidRegion())
                    return;

                if(!rubberBandFlg)
                    CreateRubberBandShapesAsync();

                return;
            }
        }
    }
}
